////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2018 The University of North Carolina at Chapel Hill.
// All Rights Reserved.
//
// Permission to use, copy, modify and distribute this software and its
// documentation for educational, research and non-profit purposes, without
// fee, and without a written agreement is hereby granted, provided that the
// above copyright notice and the following three paragraphs appear in all
// copies.
//
// IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL BE
// LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR
// CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE
// USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY
// OF NORTH CAROLINA HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGES.
//
// THE UNIVERSITY OF NORTH CAROLINA SPECIFICALLY DISCLAIM ANY
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
// PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
// NORTH CAROLINA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
// UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// The authors may be contacted via:
//
// US Mail: Real-Time Systems Group at UNC
// Department of Computer Science
// Sitterson Hall
// University of N. Carolina
// Chapel Hill, NC 27599-3175
//
// EMail: nemitz@cs.unc.edu; tamert@cs.unc.edu; anderson@cs.unc.edu
//
////////////////////////////////////////////////////////////////////////////////

#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>

#include "cycles.h"
#include "crnlp_uniform_flls.h"
#include "nesting.h"
#include "mem.h"
#include "queue.h"

void crnlp_uniform_flls_init(crnlp_uniform_flls_lock_t* lock)
{
    int i, j;

    for (i = 0; i < NUM_SERVERS; i++)
    {
        for (j=0; j < SCHEDSIZE; j++)    { lock->servers[i].enabled[i] = 0; }
        for (j=0; j < SCHEDSIZE; j++)    { lock->servers[i].blocked[i] = 0; }
        for (j=0; j < SCHEDSIZE; j++)    { lock->servers[i].schedule[i] = 0; }
        for (j=0; j < QUEUE_LENGTH; j++) { lock->servers[i].requests[i].done = 0;}
        for (j=0; j < QUEUE_LENGTH; j++) { lock->servers[i].requests[i].needs_service = 0;}
        for (j=0; j < QUEUE_LENGTH; j++) { lock->servers[i].requests[i].spin_location = -1;} // invalid

        lock->servers[i].pending_requests = 0;

        // Initialize the RR lock logic components
        lock->servers[i].rr_wave = 0;

        lock->servers[i].service_index = 0;
        lock->servers[i].server_exists = 0;
    }

    // Initialize the RR lock logic components
    lock->S = 0;
}

uint64_t *crnlp_uniform_flls_lock(crnlp_uniform_flls_lock_t* lock, uint8_t ind)
{
    crnlp_uniform_flls_server_lock_t *server_lock;
    crnlp_uniform_flls_node_t *curr_node, *self, *other;
    uint8_t start_in_sched, i, server_id;
    uint8_t is_lock_server;
    unsigned int other_server_bits;

    uint64_t *ret;
    uint64_t start, end;

    ret = (uint64_t *)malloc(sizeof(uint64_t) * 2);
    ret[0] = ret[1] = 0;

    start = get_cycles();

    server_id = ind / (QUEUE_LENGTH / NUM_SERVERS);
    server_lock = &(lock->servers[server_id]);
    self = &(server_lock->requests[ind]);

    // Mark this request as needing to be serviced for a lock
    self->done = 0;
    mb();
    self->needs_service = 1;

    // Check whether a lock server exists (and potentially become it)
    is_lock_server = 0;
    if (!server_lock->server_exists)
    {
        is_lock_server = __sync_bool_compare_and_swap(&(server_lock->server_exists), 0, 1);
    }

    // Wait until this request is serviced, or this core becomes the lock server
    // (if it is already the lock server, skip this; it will service itself later)
    while (!is_lock_server && self->needs_service != 0)
    {
        // Try to become the lock server (in the case the previous lock server
        // completed without servicing this request)
        if (!server_lock->server_exists)
        {
            is_lock_server = __sync_bool_compare_and_swap(&(server_lock->server_exists), 0, 1);
        }
    }

    // If this core is not the lock server, then we switch to measuring
    // blocking time (the request must have been serviced)
    if (!is_lock_server)
    {
        end = get_cycles();
        ret[0] += end - start;
        start = get_cycles();
    }

    // Wait until this request is satisfied, or this core becomes the lock server
    // (if it is already the lock server, skip this; it will eventually become
    // satisfied, after potentially servicing other requests)
    while (!is_lock_server && server_lock->enabled[self->spin_location] != 1) 
    {
        // Try to become the lock server (in order to help service requests
        // while waiting to be satisfied)
        if (!server_lock->server_exists)
        {
            is_lock_server = __sync_bool_compare_and_swap(&(server_lock->server_exists), 0, 1);
        }
    }

    // If this core is not the lock server, then the request must be satisfied
    if (!is_lock_server)
    {
        end = get_cycles();
        ret[1] += end - start;

        return ret;
    }

    // Otherwise, help out as long as this request has not yet been serviced
    // and is not yet satisfied
    while (self->needs_service != 0 || server_lock->enabled[self->spin_location] != 1)
    {
        // Check if the next request needs to be serviced
        i = server_lock->service_index;
        if (server_lock->requests[i].needs_service == 1)
        {
            // Service request i
            other = &(server_lock->requests[i]);
            if (other->done == 0)
            {
                // Lock code -- determine which row in the table in which
                // to place this request
                if (server_lock->pending_requests > 0)
                {
                    start_in_sched = (server_lock->head + 1) % SCHEDSIZE;
                    while ((server_lock->schedule[start_in_sched] & other->requested) != 0)
                    {
                        start_in_sched = (start_in_sched + 1) % SCHEDSIZE;
                    }
                }
                else
                {
                    start_in_sched = server_lock->head;
                }

                // Set the next value for the request
                other->next = (start_in_sched + 1) % SCHEDSIZE;

                // Update blocked for the next row, the count of pending
                // requests, and set the bits in this row corresponding to the
                // resources needed by the request
                server_lock->blocked[other->next] += 1;
                server_lock->pending_requests += 1;
                server_lock->schedule[start_in_sched] |= other->requested;

                // Set the spin location for the request to be its row in the
                // table; it will be satisfied when enabled is set to true
                // for this row
                other->spin_location = start_in_sched;

                mb();

                // If the server was idle, update the phase and mark the
                // server present
                if (server_lock->pending_requests == 1)
                {
                    server_lock->rr_wave = ~(server_lock->rr_wave);
                    if (server_id == 0)
                    {
                        other_server_bits = __sync_fetch_and_add(
                                                &lock->S,
                                                APRES | (server_lock->rr_wave & APHID));
                        other_server_bits = other_server_bits & BBITS;
                    }
                    else
                    {
                        other_server_bits = __sync_fetch_and_add(
                                                &lock->S,
                                                BPRES | (server_lock->rr_wave & BPHID));
                        other_server_bits = other_server_bits & ABITS;
                    }
                }

                // Mark the request as serviced
                other->needs_service = 0;
            }
            else
            {
                // Unlock code - clear the bits in the table corresponding to
                // the resources that were needed for the request
                server_lock->schedule[server_lock->head] &= ~(other->requested);

                // Decrement blocked and the pending request count, as the
                // request is now finished
                server_lock->blocked[other->next] -= 1;
                server_lock->pending_requests -= 1;

                // If this was the last request blocking the next row, then
                // update the head pointer and disable the row
                if (server_lock->blocked[other->next] == 0)
                {
                    server_lock->enabled[server_lock->head] = 0;
                    server_lock->head = other->next;
                }

                mb();

                // Mark the request as serviced
                other->needs_service = 0;

                if (server_lock->blocked[other->next] == 0)
                {
                    // Clear the XBITS for this server, allowing another
                    // lock server to become enabled
                    if (server_id == 0)
                    {
                        __sync_fetch_and_and(&lock->S, ~ABITS);
                    }
                    else
                    {
                        __sync_fetch_and_and(&lock->S, ~BBITS);
                    }

                    // If there are still requests on this server, update
                    // the wave and present bits
                    if (server_lock->pending_requests > 0)
                    {
                        server_lock->rr_wave = ~(server_lock->rr_wave);
                        if (server_id == 0)
                        {
                            other_server_bits = __sync_fetch_and_add(
                                                    &lock->S,
                                                    APRES | (server_lock->rr_wave & APHID));
                            other_server_bits = other_server_bits & BBITS;
                        }
                        else
                        {
                            other_server_bits = __sync_fetch_and_add(
                                                    &lock->S,
                                                    BPRES | (server_lock->rr_wave & BPHID));
                            other_server_bits = other_server_bits & ABITS;
                        }
                    }
                }
            }

            // If this request was just serviced, then we switch to measuring
            // blocking time
            if (i == ind)
            {
                end = get_cycles();
                ret[0] += end - start;
                start = get_cycles();
            }
        }

        // If the head is not enabled, check the RR logic to determine if
        // this lock server should be enabled
        if ((server_lock->pending_requests > 0) &&
            (server_lock->enabled[server_lock->head] != 1))
        {
            if ((server_id == 0) &&
                ((other_server_bits == 0) ||
                 (other_server_bits != (lock->S & BBITS))))
            {
                server_lock->enabled[server_lock->head] = 1;
            }
            else if ((server_id == 1) &&
                ((other_server_bits == 0) ||
                 (other_server_bits != (lock->S & ABITS))))
            {
                server_lock->enabled[server_lock->head] = 1;
            }
        }

        // Move onto the next request
        server_lock->service_index = (i + 1) % QUEUE_LENGTH;
    }

    // Set the server to not exist
    server_lock->server_exists = 0;

    end = get_cycles();
    ret[1] += end - start;
    start = get_cycles();

    // Satisfied!
    return ret;
}

uint64_t *crnlp_uniform_flls_unlock(crnlp_uniform_flls_lock_t* lock, uint8_t ind)
{
    crnlp_uniform_flls_server_lock_t *server_lock;
    crnlp_uniform_flls_node_t *curr_node, *self;
    uint8_t server_id;
    uint8_t is_lock_server;
    unsigned int other_server_bits;

    uint64_t *ret;
    uint64_t start, end;

    ret = (uint64_t *)malloc(sizeof(uint64_t) * 2);
    ret[0] = ret[1] = 0;

    start = get_cycles();

    server_id = ind / (QUEUE_LENGTH / NUM_SERVERS);
    server_lock = &(lock->servers[server_id]);
    self = &(server_lock->requests[ind]);

    // Mark this request as needing to be serviced for an unlock
    self->done = 1;
    mb();
    self->needs_service = 1;

    // Check whether a lock server exists (and potentially become it)
    is_lock_server = 0;
    if (!server_lock->server_exists)
    {
        is_lock_server = __sync_bool_compare_and_swap(&(server_lock->server_exists), 0, 1);
    }

    // Wait until this request is serviced, or this core becomes the lock server
    // (if it is already the lock server, skip this; it will service itself later)
    while (!is_lock_server && self->needs_service != 0)
    {
        // Try to become the lock server (in the case the previous lock server
        // completed without servicing this request)
        if (!server_lock->server_exists)
        {
            is_lock_server = __sync_bool_compare_and_swap(&(server_lock->server_exists), 0, 1);
        }
    }

    // If this core is the lock server and has not been serviced, be
    // greedy and just service this unlock
    if (is_lock_server && self->needs_service == 0)
    {
        // Oops, this core shouldn't be the server
        server_lock->server_exists = 0;
    }
    else if (is_lock_server)
    {
        // Clear the bits in the table
        // corresponding to the resources that were needed for this request
        server_lock->schedule[server_lock->head] &= ~(self->requested);

        // Decrement blocked and the pending request count, as this
        // request is now finished
        server_lock->blocked[self->next] -= 1;
        server_lock->pending_requests -= 1;

        // If this was the last request blocking the next row, then
        // update the head pointer and disable the row
        if (server_lock->blocked[self->next] == 0)
        {
            server_lock->enabled[server_lock->head] = 0;
            server_lock->head = self->next;
        }

        mb();

        // Mark this request as serviced
        self->needs_service = 0;

        if (server_lock->blocked[self->next] == 0)
        {
            // Clear the XBITS for this server, allowing another
            // lock server to become enabled
            if (server_id == 0)
            {
                __sync_fetch_and_and(&lock->S, ~ABITS);
            }
            else
            {
                __sync_fetch_and_and(&lock->S, ~BBITS);
            }

            // If there are still requests on this server, update
            // the wave and present bits
            if (server_lock->pending_requests > 0)
            {
                server_lock->rr_wave = ~(server_lock->rr_wave);
                if (server_id == 0)
                {
                    other_server_bits = __sync_fetch_and_add(
                                            &lock->S,
                                            APRES | (server_lock->rr_wave & APHID));
                    other_server_bits = other_server_bits & BBITS;
                }
                else
                {
                    other_server_bits = __sync_fetch_and_add(
                                            &lock->S,
                                            BPRES | (server_lock->rr_wave & BPHID));
                    other_server_bits = other_server_bits & ABITS;
                }
            }
        }

        // Set the server to not exist
        server_lock->server_exists = 0;
    }

    end = get_cycles();
    ret[0] += end - start;

    return ret;
}

