#!/usr/bin/env python
import sys
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import numpy as np
from csv import DictReader
from copy import copy
from collections import defaultdict
import argparse
import string

xlabels = {"percentnested": "Percent Nested",
           "numrequested": "Level of Nestedness",
           "cslength": "CS Length",
           "numcores": "Number of Tasks",
           "percentreads": "Percent Reads",
           "numresources": "Number of Resources"}

#LINE_STYLE = ['', 'y:+', 'm-.*', 'g-s', # 1, 2, 3
#              'g-^', 'g-o',             # 4, 5
#              'y--^', 'y--o',           # 6, 7
#              'r-x',                    # 8
#              'c-^', 'c-o',             # 9, 10
#              'r--^', 'r--o',           # 11, 12
#              'b-.^',                   # 13
#              'c--', 'g--', 'y--',      # 14, 15, 16 (k=4,5,6)
#              'r--', 'm--', 'k--',      # 17, 18, 19 (k=7,8,9)
#              'b--', 'c-x']             # 20, 21     (k=10,2)
LINE_STYLE = ['', 'y:+', 'm-.*', 'g-s', # 1, 2, 3
              'g-^', 'g-o',             # 4, 5
              'y--^', 'y--o',           # 6, 7
              'r-x',                    # 8
              'c-^', 'c-o',             # 9, 10
              'r--^', 'r--o',           # 11, 12
              'b-.^',                   # 13
              'b-.^', 'g--', 'y--',      # 14, 15, 16 (k=4,5,6)
              'r--', 'm--', 'k--',      # 17, 18, 19 (k=7,8,9)
              'b--', 'c-x']             # 20, 21     (k=10,2)

labels=['--', 'MCS', 'RNLP', 'U-C-RNLP',       # 1, 2, 3
        'U-C-RNLP + SGLS', 'U-C-RNLP + SLLS',  # 4, 5
        'U-C-RNLP + FGLS', 'U-C-RNLP + FLLS',  # 6, 7
        'G-C-RNLP',                            # 8
        'G-C-RNLP + SGLS', 'G-C-RNLP + SLLS',  # 9, 10
        'G-C-RNLP + FGLS', 'G-C-RNLP + FLLS',  # 11, 12
        'R3LP',                                # 13
        'CGLP', 'R5LP', 'R6LP', 'R7LP',        # 14, 15, 16, 17
        'R8LP', 'R9LP', 'R10LP', 'R2LP']       # 18, 19, 20, 21
labels2=['0','1','2','3','4','5','6','7','8','9','10','11','12','13','14','15']

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-d', "--directory", type = str,
                        default = "",
                        help = "Directory to put plot results")
    parser.add_argument('--out', type = str,
                        default = "",
                        help = "Name of output file")
    parser.add_argument('-u', "--unit", type = str,
                        default = "microseconds",
                        help = "Time unit")
    parser.add_argument('-x', "--xlabel", type = str,
                        default = "",
                        help = "X label name")
    parser.add_argument('--numcores', type = int,
                        default = -1,
                        help = "Value of numcores to whitelist")
    parser.add_argument('--numresources', type = int,
                        default = -1,
                        help = "Value of numresources to whitelist")
    parser.add_argument('--numrequested', type = int,
                        default = -1,
                        help = "Value of numrequested to whitelist")
    parser.add_argument('--cslength', type = int,
                        default = -1,
                        help = "Value of cslength to whitelist")
    parser.add_argument('--trackvars', type = str,
                        default = "",
                        help = "Comma-separated list of variant fields")
    parser.add_argument("--field", type = str,
                        help = "Field name")
    parser.add_argument("--file", type=argparse.FileType('r'), nargs='+',
                        help = "Files of csv as input")
    args = parser.parse_args()

    trackvars = args.trackvars.split(',')

    plt.figure(figsize=(10,5))
    for label in ['blocking', 'lock', 'unlock']:
        if args.trackvars == '':
            plt.figure(figsize=(10,5))

        plotvals = defaultdict(list)

        i = 0
        for file in args.file:
            #print file.name
            #f = open(file, 'r')
            d = DictReader(file)
            data = defaultdict(list)

            # Keep one data dictionary for each variation
            datavars = {}
            allvals = []
            for tv in trackvars:
                if tv == "numrequested":
                    allvals.append([1,2,4,6,8,10])
                elif tv == "cslength":
                    allvals.append([1,20,40,60,80,100])
                elif tv == "numcores":
                    allvals.append([18,34])
                elif tv == "numresources":
                    allvals.append([16,32,64])

            #print trackvars

            if len(trackvars) == 1 and trackvars[0] != '':
                for val in allvals[0]:
                    datavars[(val,)] = defaultdict(list)

            elif len(trackvars) == 2:
                for val1 in allvals[0]:
                    for val2 in allvals[1]:
                        datavars[(val1, val2)] = defaultdict(list)

            elif len(trackvars) == 3:
                for val1 in allvals[0]:
                    for val2 in allvals[1]:
                        for val3 in allvals[2]:
                            datavars[(val1, val2, val3)] = defaultdict(list)

            elif len(trackvars) == 4:
                for val1 in allvals[0]:
                    for val2 in allvals[1]:
                        for val3 in allvals[2]:
                            for val4 in allvals[3]:
                                datavars[(val1, val2, val3, val4)] = defaultdict(list)

            for row in d:

                # Filter data by keeping some numcores constant
                if args.numcores >= 0:
                    #print row["numcores"]
                    if row["numcores"] != str(args.numcores):
                        continue

                # Filter data by keeping numresources constant
                if args.numresources >= 0:
                    #print row["numresources"]
                    if row["numresources"] != str(args.numresources):
                        continue

                # Filter data by keeping numrequested constant
                if args.numrequested >= 0:
                    #print row["numrequested"]
                    if int(row["numrequested"]) != int(args.numrequested):
                        continue

                # Filter data by keeping cslength constant
                if args.cslength >= 0:
                    #print row["cslength"]
                    if row["cslength"] != str(args.cslength):
                        continue

                for key, value in row.iteritems():
                    data[key].append(value)

                passedTrackFilter = True
                if args.trackvars != '':
                    for tv_idx in range(len(trackvars)):
                        tv = trackvars[tv_idx]
                        rowval = row[tv]

                        if int(rowval) not in allvals[tv_idx]:
                            passedTrackFilter = False
                            break

                if not passedTrackFilter:
                    continue

                for key, value in row.iteritems():
                    if args.trackvars != "":
                        if len(trackvars) == 1:
                            tup = (int(row[trackvars[0]]), )
                        elif len(trackvars) == 2:
                            tup = (int(row[trackvars[0]]), int(row[trackvars[1]]))
                        elif len(trackvars) == 3:
                            tup = (int(row[trackvars[0]]), int(row[trackvars[1]]), int(row[trackvars[2]]))
                        elif len(trackvars) == 4:
                            tup = (int(row[trackvars[0]]), int(row[trackvars[1]]), int(row[trackvars[2]]),
                                   int(row[trackvars[3]]))
                        datavars[tup][key].append(float(value))
                    #else:
                    #    data[key].append(float(value))

            file.seek(0)

            #print "Found %d data points" % len(data["numcores"])

            if args.trackvars != "":
                indx = string.atoi(file.name[14:file.name.index(".")])
            else:
                underscoreIndx = file.name.index("_")
                dotIndx = file.name.index(".", underscoreIndx)
                indx = string.atoi(file.name[underscoreIndx+1:dotIndx])

            if args.trackvars != "":
                for key in datavars:
                    if len(datavars[key][label]) > 0:
                        x,y = zip(*sorted(zip(datavars[key][args.field], datavars[key][label]), key=lambda x:x[0]))
                        plotvals[key].append((x,y,indx))
            else:
                x,y = zip(*sorted(zip(data[args.field], data[label]), key=lambda x:float(x[0])))
                s = LINE_STYLE[indx]
                if s[0] == 'o':
                    plt.plot(x, y, color="#FFB07C", marker=s[-1], linestyle=s[1:-1], label=labels[indx], linewidth=2.0)
                else:
                    #plt.plot(x, y, s, label=labels[indx], linewidth=2.0)

                    x,y1 = zip(*sorted(zip(data[args.field], data["lock"]), key=lambda x:int(x[0])))       
                    x,y2 = zip(*sorted(zip(data[args.field], data["unlock"]), key=lambda x:int(x[0])))       
                    y = [float(y1[i]) + float(y2[i]) for i in range(len(x))]
                    #y = [float(y1[i][1]) + float(y2[i][1]) for i in range(len(y1))]
                    plt.plot(x, y, s, label=labels[indx], linewidth=2.0)

        if args.trackvars != "":
            for key in plotvals:
                for (x,y,indx) in plotvals[key]:
                    s = LINE_STYLE[indx]
                    if s[0] == 'o':
                        plt.plot(x, y, color="#FFB07C", marker=s[-1], linestyle=s[1:-1], label=labels[indx], linewidth=2.0)
                    else:
                        plt.plot(x, y, s, label=labels[indx], linewidth=2.0)

                    plt.legend(loc="best")

                if label=='blocking':
                    plt.ylabel("Blocking ("+args.unit+")")
                elif label=='lock':
                    plt.ylabel("Lock Overhead ("+args.unit+")")
                    xmin,xmax,ymin,ymax = plt.axis()
                elif label=='unlock':
                    plt.ylabel("Unlock Overhead ("+args.unit+")")
                    xmin,xmax,ymin,ymax = plt.axis()
                if args.field in xlabels:
                    plt.xlabel(xlabels[args.field])
                else:
                    plt.xlabel(args.xlabel)

                namevals = []
                for param in ["numcores","numresources","numrequested","cslength"]:
                    if param == args.field:
                        namevals.append("x")
                    elif param in trackvars:
                        namevals.append(str(key[trackvars.index(param)]))
                namestr = "-".join(namevals)

                filename = args.out if args.out != "" else label
                plt.savefig(args.directory+"/"+args.field+"/"+filename+"_"+namestr+".png")
                plt.savefig(args.directory+"/"+args.field+"/"+filename+"_"+namestr+".pdf")

                plt.clf()

        else:
            plt.legend(loc="best")
            #plt.legend(loc="center left")

            if label=='blocking':
                plt.ylabel("Blocking ("+args.unit+")")
            elif label=='lock':
                #plt.ylabel("Lock Overhead ("+args.unit+")")
                plt.ylabel("Overhead ("+args.unit+")")
            elif label=='unlock':
                #plt.ylabel("Unlock Overhead ("+args.unit+")")
                plt.ylabel("Overhead ("+args.unit+")")

            if args.field in xlabels:
                plt.xlabel(xlabels[args.field])
            else:
                plt.xlabel(args.xlabel)
            filename = "%s_%s" % (args.out, label) if args.out != "" else label
            plt.savefig(args.directory+"/"+filename+".png")
            plt.savefig(args.directory+"/"+filename+".pdf")


#            alldata = zip(data["numcores"], data[label])
#            x, y = zip(*sorted(alldata, key=lambda x:int(x[0])))
#
#            s = LINE_STYLE[i]
#            i = i + 1
#            plt.plot(x, y, s, label=file[6:-4], linewidth=2.0)
#
#        plt.legend(loc="upper left")
#        #    plt.legend(loc="best")
#        if label == "blocking":
#            plt.ylabel("Blocking (microseconds)")
#        elif label == "lock":
#            plt.ylabel("Lock Overhead (microseconds)")
#        if label == "unlock":
#            plt.ylabel("Unlock Overhead (microseconds)")
#        plt.xlabel("Tasks")
#
#        plt.savefig(file[:5]+"_"+label+".pdf")
#        plt.savefig(file[:5]+"_"+label+".png")

if __name__ == '__main__':
    main()
