#!/usr/bin/env python

from __future__ import division

import argparse
import random
import sys
import math
import time
import traceback
import multiprocessing
import multiprocessing.pool

from collections import defaultdict
from csv import DictWriter
from itertools import product
from math import ceil
from multiprocessing import cpu_count
from numpy import arange
from pprint import pprint
from copy import deepcopy
#from timeout_decorator import timeout

from schedcat.model.tasks import SporadicTask, TaskSystem
from schedcat.generator.tasks import exponential, uniform, multimodal
from schedcat.generator.tasksets import NAMED_PERIODS, NAMED_UTILIZATIONS
from schedcat.util.storage import storage
import schedcat.sched.edf as edf
import schedcat.model.resources as resources
import schedcat.generator.tasks as tasks
import schedcat.locking.bounds as bounds
from schedcat.util.time import ms2us, ms27s, ms28s, ms2ns

from util.stats import BernoulliEstimator

MIN_SAMPLES = 150#1000
MAX_SAMPLES = 200#2000
MAX_CI      = 0.05
CONFIDENCE  = 0.95
convert_func = ms2us
use_overhead = False

#microseconds
if (convert_func == ms2us):

    # RTNS'19 overhead values
    O_UCRNLP  = 13
    O_GCRNLP  = 15

    O_CGLP_ARR  = [0, 0, 3, 3, 3, 3, 3, 3, 3, 3, 3]
    # for k of...     1  2  3  4  5  6  7  8  9 10

    NAMED_CS_LENGTHS = { # In micro seconds
        'short'         : uniform(1, 15),
        'moderate'      : uniform(15, 100),
        'xsmall'        : exponential(1,    100,    10),
        'small'         : exponential(10,   1000,   100),
        'large'         : exponential(100,  10000,  1000),
        'long'          : uniform(100, 1000),
        'bimo-medium'   : multimodal([(uniform(15,   100), 5),
                                      (uniform(500, 1000), 5)]),
        'bimo-medium2'  : multimodal([(uniform(15,   100), 7),
                                      (uniform(500, 1000), 3)]),
        'bimo-moderate' : multimodal([(uniform(15,   100), 6),
                                      (uniform(100, 1000), 3)]),
        'bimo-long'     : multimodal([(uniform(15,   100), 4),
                                      (uniform(100, 1000), 5)])
    }
elif (convert_func == ms27s):
# -7-sec

    # RTNS'19 overhead values
    O_UCRNLP  = 128
    O_GCRNLP  = 150

    O_CGLP_ARR  = [0, 0, 25, 27, 28, 30, 28, 28, 28, 28, 28]
    # for k of...     1   2   3   4   5   6   7   8   9  10

    NAMED_CS_LENGTHS = { # In -7 seconds
        'short'         : uniform(10, 150),
        'moderate'      : uniform(150, 1000),
        'xsmall'        : exponential(10,    1000,    100),
        'small'         : exponential(100,   10000,   1000),
        'large'         : exponential(1000,  100000,  10000),
        'long'          : uniform(1000, 10000),
        'bimo-medium'   : multimodal([(uniform(150,   1000), 5),
                                      (uniform(5000, 10000), 5)]),
        'bimo-medium2'  : multimodal([(uniform(150,   1000), 7),
                                      (uniform(5000, 10000), 3)]),
        'bimo-moderate' : multimodal([(uniform(150,   1000), 6),
                                      (uniform(1000, 10000), 3)]),
        'bimo-long'     : multimodal([(uniform(150,   1000), 4),
                                      (uniform(1000, 10000), 5)])
    }
elif (convert_func == ms28s):
# e-8 seconds

    # RTNS'19 overhead values
    O_UCRNLP  = 1276
    O_GCRNLP  = 1502

    O_CGLP_ARR  = [0, 0, 251, 271, 276, 300, 277, 281, 281, 280, 279]
    # for k of...     1    2    3    4    5    6    7    8    9   10

    NAMED_CS_LENGTHS = { # In nano seconds
        'short'         : uniform(100, 1500),
        'moderate'      : uniform(1500, 10000),
        'xsmall'        : exponential(100,    10000,    1000),
        'small'         : exponential(1000,   100000,   10000),
        'large'         : exponential(10000,  1000000,  100000),
        'long'          : uniform(10000, 100000),
        'bimo-medium'   : multimodal([(uniform(1500,   10000), 5),
                                      (uniform(50000, 100000), 5)]),
        'bimo-medium2'  : multimodal([(uniform(1500,   10000), 7),
                                      (uniform(50000, 100000), 3)]),
        'bimo-moderate' : multimodal([(uniform(1500,   10000), 6),
                                      (uniform(10000, 100000), 3)]),
        'bimo-long'     : multimodal([(uniform(1500,   10000), 4),
                                      (uniform(10000, 100000), 5)])
    }
elif (convert_func == ms2ns):
# nanoseconds

    # RTNS'19 overhead values
    O_UCRNLP  = 12764
    O_GCRNLP  = 15019

    O_CGLP_ARR  = [0, 0, 2510, 2710, 2757, 2998, 2767, 2811, 2806, 2798, 2793]
    # for k of...     1     2     3     4     5     6     7     8     9    10

    NAMED_CS_LENGTHS = { # In nano seconds
        'short'         : uniform(1000, 15000),
        'moderate'      : uniform(15000, 100000),
        'xsmall'        : exponential(1000,    100000,    10000),
        'small'         : exponential(10000,   1000000,   100000),
        'large'         : exponential(100000,  10000000,  1000000),
        'long'          : uniform(100000, 1000000),
        'bimo-medium'   : multimodal([(uniform(15000,   100000), 5),
                                      (uniform(500000, 1000000), 5)]),
        'bimo-medium2'  : multimodal([(uniform(15000,   100000), 7),
                                      (uniform(500000, 1000000), 3)]),
        'bimo-moderate' : multimodal([(uniform(15000,   100000), 6),
                                      (uniform(100000, 1000000), 3)]),
        'bimo-long'     : multimodal([(uniform(15000,   100000), 4),
                                      (uniform(100000, 1000000), 5)])
    }

def random_is_read(prob):
    return random.random() < prob

def random_is_nested(prob):
    return random.random() < prob

def get_resource(total):
    return random.randint(0,total-1)

def get_resources(total, needed):
    lst = []
    while len(lst) < needed:
        temp = random.randint(0,total-1)
        if temp not in lst:
            lst.append(temp)
    return lst

def determine_group_count(numRequests, nestedProb, numResources,
                          numResourcesNested, numTasks):
    num_resources_requested = int(numRequests * nestedProb * numResourcesNested + \
                                  numRequests * (1 - nestedProb))
    num_groups_min = max(2, int(math.ceil(num_resources_requested / numResources)))
    return num_groups_min

def create_group_resources(numGroups, numResources, cslenFunc):
    """
    Returns a list of groups, where each group is represented by the
    unclaimed resources for that group.
    """
    resources_by_group = []

    for i in range(numGroups):
        cs_len = int(NAMED_CS_LENGTHS[cslenFunc]())
        resources_by_group.append( (range(numResources), cs_len) )

    return resources_by_group

def get_group_resources(resourcesByGroup, numRequesting, numResources, cslenFunc):
    """
    Randomly chooses a group and numRequesting number of resources from that
    group.  Returns the group number and the list of selected resource numbers.
    """
    group_has_space = []
    for group_num in range(len(resourcesByGroup)):
        if len(resourcesByGroup[group_num][0]) >= numRequesting:
            group_has_space.append(group_num)

    # If there aren't any groups with space, make a new one
    if len(group_has_space) > 0:
        group_num = random.choice(group_has_space)
    else:
        group_num = len(resourcesByGroup)
        group_cs_len = int(NAMED_CS_LENGTHS[cslenFunc]())
        resourcesByGroup.append( (range(numResources), group_cs_len) )

    resources_chosen = []
    for i in range(numRequesting):
        idx = random.randint(0, len(resourcesByGroup[group_num][0])-1)
        resources_chosen.append(resourcesByGroup[group_num][0].pop(idx))

    cs_len = int(random.uniform(0.8, 1.0) * resourcesByGroup[group_num][1])
    #cs_len = resourcesByGroup[group_num][1]

    return group_num, resources_chosen, cs_len

# the following two classes are take from 
# http://stackoverflow.com/questions/6974695/python-process-pool-non-daemonic
# without time, you cannot use timeouts within the threads.
class NoDaemonProcess(multiprocessing.Process):
    def _get_daemon(self):
        return False
    def _set_daemon(self, value):
        pass
    daemon = property(_get_daemon, _set_daemon)

class MyPool(multiprocessing.pool.Pool):
    Process = NoDaemonProcess

class ExperimentManager(object):
    def __init__(self, parameters, outfile):
        random.seed(12345)
        self.parameters = parameters
        self.outfile = outfile
        self.out = DictWriter(outfile, parameters.keys()+DesignPoint.metrics)
        self.out.writeheader()
        self.create_design_points()
    
    def valid(self, dp):
        return dp.processors >= dp.sys_util

    def create_design_points(self, shuffle = True):
        self.design_points = filter(self.valid,
            [DesignPoint(**dict(zip(self.parameters.keys(),val)))
                for val in product(*self.parameters.values())])

        if shuffle:
            random.shuffle(self.design_points)

    def run(self, processors, status = True):
        if processors == 1:
            self._run_serial(status)
        else:
            self._run_parallel(processors, status)

        self.finish()

        return # need to return the results here.

    #Useful for debugging purposes.
    def _run_serial(self, status = True):
        rows = map(lambda dp: dp.run(), self.design_points)
        for i, row in enumerate(rows):
            if status and (self.outfile != sys.stdout):
                sys.stderr.write('\rdone {0:%}'.format(i/len(rows)))
            self.out.writerow(row)

    def _run_parallel(self, processors, status = True):
        #pool = Pool(processes = processors)
        pool = MyPool(processes = processors)
        try:
            for i, row in enumerate(pool.imap_unordered(pickleable_function, self.design_points)):
                if status and (self.outfile != sys.stdout):
                    sys.stderr.write('\rdone {0:%}'.format(i/len(self.design_points)))
                self.out.writerow(row)
            pool.close()
        except Exception as e:
            pool.terminate()
            print e
            raise

    def finish(self):
        self.outfile.close()

# for whatever reason, imap_unordered will not take a lambda function, because
# it is not pickleable.
def pickleable_function(dp):
    return dp.run()

class DesignPoint(object):
    metrics = ["UCRNLP","GCRNLP", "CGLP", "NOLOCK"]

    # a graph which represents which tests can decide a metric on a given run.
    test_graph = {}

    def __init__(self, **levels):
        self.__dict__.update(levels)
        self.levels = levels
        self.data = dict([(m, BernoulliEstimator()) for m in DesignPoint.metrics])

    def run(self):
        while not self.complete():
            ucrnlp_ts, gcrnlp_ts, cglp_ts, org_ts = self.create_task_set()

            # Write some logic to run all your tests here. This is often study
            # specific, as you can exploit domincance results to accelerate
            # things. A classic example, if the task system isn't schedulable
            # without any synchronization, it isn't schedulable with it.
            if edf.is_schedulable(self.processors, org_ts):
                self.data["NOLOCK"].add_sample(True)

                self.ucrnlp(ucrnlp_ts)
                self.gcrnlp(gcrnlp_ts)
                self.cglp(cglp_ts)
            else:
                for m in DesignPoint.metrics:
                    self.data[m].add_sample(False)

        return dict(self.levels.items() + [(k,v.mean) for (k,v) in self.data.iteritems()])

    def ucrnlp(self, ts):
        #@timeout(10, use_signals = False)
        def test():
            bounds.apply_ucrnlp_bounds(ts, self.processors)
            return edf.is_schedulable(self.processors, ts,
                                     want_baruah = True,
                                     want_rta    = False,
                                     want_ffdbf  = False,
                                     want_load   = False)
        #try:
        res = test()
        #except:
        #    res = False
        self.data["UCRNLP"].add_sample(res)
        return res

    def gcrnlp(self, ts):
        #@timeout(10, use_signals = False)
        def test():
            bounds.apply_gcrnlp_bounds(ts, self.processors)
            return edf.is_schedulable(self.processors, ts,
                                     want_baruah = True,
                                     want_rta    = False,
                                     want_ffdbf  = False,
                                     want_load   = False)
        #try:
        res = test()
        #except:
        #    res = False
        self.data["GCRNLP"].add_sample(res)
        return res

    def cglp(self, ts):
        #@timeout(10, use_signals = False)
        def test():
            bounds.apply_cglp_bounds(ts, self.processors)
            return edf.is_schedulable(self.processors, ts,
                                     want_baruah = True,
                                     want_rta    = False,
                                     want_ffdbf  = False,
                                     want_load   = False)
        #try:
        res = test()
        #except:
        #    res = False
        self.data["CGLP"].add_sample(res)
        return res

    def complete(self):
        for metric, stat in self.data.iteritems():
            if stat.count < MIN_SAMPLES:
                return False
            elif stat.count > MAX_SAMPLES:
                return True
            elif stat.confidence_interval(CONFIDENCE) > MAX_CI:
                return False
        return True
        
    def create_task_set(self):
        # Generate the task set
        tg = tasks.TaskGenerator(period = NAMED_PERIODS[self.period],
                                    util   = NAMED_UTILIZATIONS[self.task_util])

        ts = tg.make_task_set(max_util = self.sys_util, squeeze = True, time_conversion=convert_func)

        resources.initialize_resource_model(ts)
        bounds.assign_edf_preemption_levels(ts)

        # Make a copy of the task set for each locking protocol being tested
        ucrnlp_ts = ts.copy()
        gcrnlp_ts = ts.copy()
        cglp_ts = ts.copy()

        # Initialize the resource groups, using the number of tasks, the
        # percent requesting, and the nested percentage and level to determine
        # how many groups are needed
        num_requests = int(self.req_perc * len(ts))
        num_groups = determine_group_count(num_requests, self.nested_prob,
                                           self.numresources,
                                           self.numresourcesnested,
                                           len(ts))
        resources_by_group = create_group_resources(num_groups,
                                                    self.numresources,
                                                    self.cs_length)

        # Randomly choose self.req_perc percentage of tasks to make requests,
        # and for each of the requesting tasks, determine its request
        # parameters
        all_reqs = []
        for i in range(0, num_requests):
            # Determine whether the request is for a read or write, nested or
            #  not, and the CS length
            is_read = random_is_read(self.read_prob)
            is_nested = random_is_nested(self.nested_prob)
            if is_nested:
                num_res = self.numresourcesnested
            else:
                num_res = 1

            # Choose resources for the request; res_list is a list of resource
            # IDs, each in the range [0, self.numresources)
            group_num, res_list, cs_len = get_group_resources(resources_by_group, num_res,
                                                              self.numresources, self.cs_length)
            cs_len = int(NAMED_CS_LENGTHS[self.cs_length]())

            all_reqs.append([res_list, cs_len, is_read, group_num])

        # Map requesting tasks to the chosen request parameters
        for i in random.sample(range(len(ts)), num_requests):
            res_list, cs_len, is_read, group_num = all_reqs.pop()

            if use_overhead:
                ucrnlp_overhead = O_UCRNLP
                gcrnlp_overhead = O_GCRNLP
                cglp_overhead   = O_CGLP_ARR[len(resources_by_group)]
            else:
                ucrnlp_overhead = 0
                gcrnlp_overhead = 0
                cglp_overhead   = 0

            # For blocking calculations for the CGLP, the "resource" is
            # actually the group to which the request belongs
            cglp_ts[i].resmodel[group_num].add_request(cs_len+cglp_overhead,is_read)

            # Update the resource model with the request information
            for res_id in res_list:
                j = res_id
                ts[i].resmodel[res_id].add_request(cs_len,is_read)
                ucrnlp_ts[i].resmodel[res_id].add_request(cs_len+ucrnlp_overhead,is_read)
                gcrnlp_ts[i].resmodel[res_id].add_request(cs_len+gcrnlp_overhead,is_read)

        for protocol_ts in [ts, ucrnlp_ts, gcrnlp_ts, cglp_ts]:
            for t in protocol_ts:
                t.partition = 0
                t.response_time = t.deadline

        return ucrnlp_ts, gcrnlp_ts, cglp_ts, ts

def output_task_set(tsi, paramstr, filenum, dp, dpi):
    group_mcs_ts, ucrnlp_ts, gcrnlp_ts, cglp_ts, pftl_ts, rwrnlp_ts, rrr_ts, star_ts, org_ts = dp.create_task_set()

    filename = "%s_tasksetid-%d" % (paramstr, tsi)

    f1 = open("tasksets-full/" + ("%03d_" % filenum) + filename + ".txt", 'w')
    f2 = open("tasksets/" + ("%03d" % filenum) + ".txt", 'w')
    f1.write("[DP %d] %s\n" % (dpi, str(dp.levels)))
    for i in range(len(cglp_ts)):
        cslen = max(cglp_ts[i].resmodel.values()[0].max_read_length,
                    cglp_ts[i].resmodel.values()[0].max_write_length)
        f1.write(str(i) + " | " + str(cglp_ts[i].resmodel.keys()[0]) + " | " + str(org_ts[i].resmodel.keys()) + " | " + str(cslen) + "\n")
        f2.write(str(org_ts[i].resmodel.keys()) + "\n")

    f1.close()
    f2.close()

    print "%d tasks" % len(cglp_ts)

# Helper function to more easily specify parameters
def myrange(start, end, inc):
    return arange(start, end+inc, inc)

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-o', "--outfile", type = argparse.FileType('w'),
                        default = sys.stdout,
                        help = "File to output csv file to")
    parser.add_argument('-p', "--pretend", action='store_true',
                        help = "Only print design point, do not execute")
    parser.add_argument('-m', "--processors", default=cpu_count(), type = int,
                        help="Number of processors to execute on")
    parser.add_argument('-q', "--quiet", action='store_true',
                        help = "Quash all progress and status updates")
    args = parser.parse_args()

    params = storage()

    params.processors = [18]
    #params.sys_util = myrange(2, 16, 0.5)
    params.sys_util = myrange(2, 16, 1)
    params.task_util = ['uni-heavy', 'uni-medium']
    #params.period = ['uni-short', 'uni-long']
    params.period = ['uni-short']
    #params.cs_length = ['moderate', 'long', 'bimo-medium', 'bimo-medium2']
    params.cs_length = ['moderate']
    #params.req_perc = [0.5, 0.8, 1]
    params.req_perc = [1]
    params.read_prob = [0]
    #params.nested_prob = [0.1, 0.2, 0.5, 0.8]
    params.nested_prob = [0.5]
    params.numresources = [64]
    #params.numresourcesnested = [2, 4]
    params.numresourcesnested = [4]

    exp = ExperimentManager(params, args.outfile)

    if args.pretend:
        num_task_sets = 50
        filenum = 1
        for dpi in range(len(exp.design_points)):
            dp = exp.design_points[dpi]

            if dp.levels["task_util"] == "uni-heavy":
                dp.levels["cs_length"] = "long"
            elif dp.levels["task_util"] == "uni-medium":
                dp.levels["cs_length"] = "bimo-medium"
            elif dp.levels["task_util"] == "uni-medlight":
                dp.levels["cs_length"] = "xsmall"

            paramstr = "cslen-%s_numcores-%d_taskutil-%s_numresources-%d_period-%s_reqperc-%s_sysutil-%s_numrequested-%d_nestedprob-%s_readprob-%s" % \
                (dp.levels["cs_length"],
                 dp.levels["processors"],
                 dp.levels["task_util"],
                 dp.levels["numresources"],
                 dp.levels["period"],
                 dp.levels["req_perc"],
                 dp.levels["sys_util"],
                 dp.levels["numresourcesnested"],
                 dp.levels["nested_prob"],
                 dp.levels["read_prob"])

            for tsi in range(num_task_sets):
                output_task_set(tsi, paramstr, filenum, dp, dpi)
                filenum += 1

    if not args.quiet:
        print "Total design points: ", len(exp.design_points)

    if not args.pretend:
        exp.run(args.processors, not args.quiet)

if __name__ == '__main__':
    main()
